"""
Tests for Firebird database features.
"""

from unittest.mock import MagicMock

import pytest


class TestDatabaseFeatures:
    """Test cases for DatabaseFeatures."""

    @pytest.fixture
    def features_25(self):
        """Features instance for Firebird 2.5."""
        from django_firebird.features import DatabaseFeatures

        mock_connection = MagicMock()
        mock_connection.firebird_version = (2, 5)
        return DatabaseFeatures(mock_connection)

    @pytest.fixture
    def features_30(self):
        """Features instance for Firebird 3.0."""
        from django_firebird.features import DatabaseFeatures

        mock_connection = MagicMock()
        mock_connection.firebird_version = (3, 0)
        return DatabaseFeatures(mock_connection)

    @pytest.fixture
    def features_40(self):
        """Features instance for Firebird 4.0."""
        from django_firebird.features import DatabaseFeatures

        mock_connection = MagicMock()
        mock_connection.firebird_version = (4, 0)
        return DatabaseFeatures(mock_connection)

    @pytest.fixture
    def features_50(self):
        """Features instance for Firebird 5.0."""
        from django_firebird.features import DatabaseFeatures

        mock_connection = MagicMock()
        mock_connection.firebird_version = (5, 0)
        return DatabaseFeatures(mock_connection)

    # =========================================================================
    # Basic features (all versions)
    # =========================================================================

    def test_supports_transactions(self, features_25):
        """All Firebird versions support transactions."""
        assert features_25.supports_transactions is True

    def test_supports_savepoints(self, features_25):
        """All Firebird versions support savepoints."""
        assert features_25.uses_savepoints is True
        assert features_25.can_release_savepoints is True

    def test_cannot_rollback_ddl(self, features_25):
        """DDL is auto-committed in Firebird."""
        assert features_25.can_rollback_ddl is False

    def test_supports_foreign_keys(self, features_25):
        """All Firebird versions support foreign keys."""
        assert features_25.supports_foreign_keys is True
        assert features_25.can_introspect_foreign_keys is True

    # =========================================================================
    # SELECT FOR UPDATE features
    # =========================================================================

    def test_has_select_for_update(self, features_25):
        """Firebird supports SELECT ... WITH LOCK."""
        assert features_25.has_select_for_update is True

    def test_has_select_for_update_nowait(self, features_25):
        """Firebird WITH LOCK has implicit nowait behavior."""
        assert features_25.has_select_for_update_nowait is True

    def test_no_select_for_update_of(self, features_25):
        """Firebird doesn't support FOR UPDATE OF clause."""
        assert features_25.has_select_for_update_of is False

    def test_skip_locked_firebird_25(self, features_25):
        """Firebird 2.5 does not support SKIP LOCKED."""
        assert features_25.has_select_for_update_skip_locked is False

    def test_skip_locked_firebird_50(self, features_50):
        """Firebird 5.0 supports SKIP LOCKED."""
        assert features_50.has_select_for_update_skip_locked is True

    # =========================================================================
    # RETURNING clause
    # =========================================================================

    def test_can_return_from_insert(self, features_25):
        """All Firebird versions support RETURNING from INSERT."""
        assert features_25.can_return_columns_from_insert is True

    def test_cannot_return_bulk_insert(self, features_25):
        """Firebird only supports single-row RETURNING."""
        assert features_25.can_return_rows_from_bulk_insert is False

    def test_can_return_from_update(self, features_25):
        """Firebird 2.1+ supports UPDATE ... RETURNING."""
        assert features_25.can_return_rows_from_update is True

    # =========================================================================
    # Index features
    # =========================================================================

    def test_supports_partial_indexes(self, features_25):
        """Firebird supports partial indexes (WHERE clause)."""
        assert features_25.supports_partial_indexes is True
        assert features_25.supports_functions_in_partial_indexes is True

    def test_supports_expression_indexes(self, features_25):
        """Firebird supports expression indexes (COMPUTED BY)."""
        assert features_25.supports_expression_indexes is True

    def test_no_covering_indexes(self, features_50):
        """Firebird doesn't support covering indexes (INCLUDE)."""
        assert features_50.supports_covering_indexes is False

    def test_cannot_rename_index(self, features_25):
        """Firebird doesn't support ALTER INDEX ... RENAME."""
        assert features_25.can_rename_index is False

    # =========================================================================
    # Boolean field (version dependent)
    # =========================================================================

    def test_no_native_boolean_firebird_25(self, features_25):
        """Firebird 2.5 has no native BOOLEAN type."""
        assert features_25.has_native_boolean_field is False

    def test_native_boolean_firebird_30(self, features_30):
        """Firebird 3.0 has native BOOLEAN type."""
        assert features_30.has_native_boolean_field is True

    # =========================================================================
    # JSON support (none)
    # =========================================================================

    def test_no_json_support(self, features_50):
        """Firebird does not have JSON support."""
        assert features_50.supports_json_field is False
        assert features_50.has_native_json_field is False
        assert features_50.has_json_operators is False
        assert features_50.can_introspect_json_field is False

    # =========================================================================
    # Timezone support (version dependent)
    # =========================================================================

    def test_no_timezone_firebird_30(self, features_30):
        """Firebird 3.0 does not support timezones."""
        assert features_30.supports_timezones is False
        assert features_30.has_zoneinfo_database is False

    def test_timezone_firebird_40(self, features_40):
        """Firebird 4.0 supports TIME/TIMESTAMP WITH TIME ZONE."""
        assert features_40.supports_timezones is True
        assert features_40.has_zoneinfo_database is True

    # =========================================================================
    # Generated columns
    # =========================================================================

    def test_supports_stored_generated_columns(self, features_25):
        """Firebird supports COMPUTED BY columns."""
        assert features_25.supports_stored_generated_columns is True

    def test_no_virtual_generated_columns(self, features_50):
        """Firebird doesn't have virtual generated columns."""
        assert features_50.supports_virtual_generated_columns is False

    # =========================================================================
    # Window functions (version dependent)
    # =========================================================================

    def test_no_window_functions_firebird_25(self, features_25):
        """Firebird 2.5 does not support window functions."""
        assert features_25.supports_over_clause is False
        assert features_25.supports_aggregate_filter_clause is False

    def test_window_functions_firebird_30(self, features_30):
        """Firebird 3.0 supports window functions."""
        assert features_30.supports_over_clause is True
        assert features_30.supports_frame_range_fixed_distance is True
        assert features_30.supports_aggregate_filter_clause is True

    def test_no_frame_exclusion(self, features_50):
        """Firebird doesn't support EXCLUDE clause in frames."""
        assert features_50.supports_frame_exclusion is False

    # =========================================================================
    # NULL handling
    # =========================================================================

    def test_nulls_order(self, features_25):
        """Firebird sorts NULLs first by default."""
        assert features_25.nulls_order_largest is False
        assert features_25.order_by_nulls_first is True

    def test_supports_nulls_modifier(self, features_25):
        """Firebird supports NULLS FIRST/LAST."""
        assert features_25.supports_order_by_nulls_modifier is True

    def test_implied_column_null(self, features_25):
        """Columns are nullable by default in Firebird."""
        assert features_25.implied_column_null is True

    # =========================================================================
    # Unsupported features
    # =========================================================================

    def test_no_tablespaces(self, features_25):
        """Firebird doesn't have tablespaces."""
        assert features_25.supports_tablespaces is False

    def test_no_distinct_on(self, features_25):
        """Firebird doesn't support DISTINCT ON."""
        assert features_25.can_distinct_on_fields is False

    def test_no_clone_databases(self, features_25):
        """Firebird doesn't have built-in database cloning."""
        assert features_25.can_clone_databases is False

    def test_no_defer_constraints(self, features_25):
        """Firebird checks constraints immediately."""
        assert features_25.can_defer_constraint_checks is False
        assert features_25.supports_deferrable_unique_constraints is False

    # =========================================================================
    # UPSERT features
    # =========================================================================

    def test_no_on_conflict_syntax(self, features_50):
        """Firebird uses UPDATE OR INSERT, not ON CONFLICT."""
        assert features_50.supports_ignore_conflicts is False
        assert features_50.supports_update_conflicts is False

    # =========================================================================
    # Version detection helpers
    # =========================================================================

    def test_is_firebird_3(self, features_25, features_30):
        """Test is_firebird_3 property."""
        assert features_25.is_firebird_3 is False
        assert features_30.is_firebird_3 is True

    def test_is_firebird_4(self, features_30, features_40):
        """Test is_firebird_4 property."""
        assert features_30.is_firebird_4 is False
        assert features_40.is_firebird_4 is True

    def test_is_firebird_5(self, features_40, features_50):
        """Test is_firebird_5 property."""
        assert features_40.is_firebird_5 is False
        assert features_50.is_firebird_5 is True

    # =========================================================================
    # Introspected field types
    # =========================================================================

    def test_introspected_field_types_firebird_25(self, features_25):
        """Firebird 2.5 maps BooleanField to SmallIntegerField."""
        types = features_25.introspected_field_types
        assert types["BooleanField"] == "SmallIntegerField"
        assert types["PositiveIntegerField"] == "IntegerField"

    def test_introspected_field_types_firebird_30(self, features_30):
        """Firebird 3.0 has native BooleanField."""
        types = features_30.introspected_field_types
        assert (
            "BooleanField" not in types or types.get("BooleanField") == "BooleanField"
        )

    # =========================================================================
    # Test skips
    # =========================================================================

    def test_django_test_skips_base(self, features_50):
        """Basic test skips are present."""
        skips = features_50.django_test_skips
        assert "Firebird does not support covering indexes" in skips
        assert "Firebird does not support DISTINCT ON" in skips
        assert "Firebird does not support JSON field" in skips

    def test_django_test_skips_version_dependent(self, features_25):
        """Version-dependent test skips for Firebird 2.5."""
        skips = features_25.django_test_skips
        assert "Firebird < 3.0 does not support window functions" in skips
        assert "Firebird < 4.0 does not support timezones" in skips
        assert "Firebird < 5.0 does not support SKIP LOCKED" in skips

    def test_no_extra_skips_firebird_50(self, features_50):
        """Firebird 5.0 doesn't have version-dependent skips."""
        skips = features_50.django_test_skips
        assert "Firebird < 3.0 does not support window functions" not in skips
        assert "Firebird < 4.0 does not support timezones" not in skips
        assert "Firebird < 5.0 does not support SKIP LOCKED" not in skips

    # =========================================================================
    # Test procedures
    # =========================================================================

    def test_create_test_procedure_sql(self, features_25):
        """Test procedure SQL is defined."""
        assert features_25.create_test_procedure_without_params_sql is not None
        assert (
            "CREATE PROCEDURE test_procedure"
            in features_25.create_test_procedure_without_params_sql
        )
        assert features_25.create_test_procedure_with_int_param_sql is not None
        assert "P_I INTEGER" in features_25.create_test_procedure_with_int_param_sql

    # =========================================================================
    # Misc features
    # =========================================================================

    def test_truncates_names(self, features_25):
        """Firebird truncates long identifiers."""
        assert features_25.truncates_names is True

    def test_no_pyformat_paramstyle(self, features_25):
        """Firebird uses ? placeholders, not %(name)s."""
        assert features_25.supports_paramstyle_pyformat is False

    def test_supports_comments(self, features_25):
        """Firebird supports COMMENT ON."""
        assert features_25.supports_comments is True
        assert features_25.supports_comments_inline is False

    def test_explain_formats(self, features_25):
        """Firebird supports TEXT explain format."""
        assert "TEXT" in features_25.supported_explain_formats

    # =========================================================================
    # Additional PostgreSQL-compatible flags
    # =========================================================================

    def test_insert_test_table_with_defaults(self, features_25):
        """Firebird supports INSERT ... DEFAULT VALUES."""
        assert features_25.insert_test_table_with_defaults is not None
        assert "DEFAULT VALUES" in features_25.insert_test_table_with_defaults

    def test_only_supports_unbounded(self, features_25):
        """Firebird 3+ supports value expressions in window frames."""
        assert features_25.only_supports_unbounded_with_preceding_and_following is False

    def test_requires_casted_case_in_updates(self, features_25):
        """Firebird doesn't require CASE casting in UPDATEs."""
        assert features_25.requires_casted_case_in_updates is False

    def test_schema_editor_clientside_binding(self, features_25):
        """Firebird doesn't use client-side param binding in schema editor."""
        assert features_25.schema_editor_uses_clientside_param_binding is False

    def test_supports_any_value(self, features_25):
        """Firebird doesn't support SQL 2023 ANY_VALUE."""
        assert features_25.supports_any_value is False

    def test_supports_nulls_distinct_unique(self, features_25):
        """Firebird doesn't support NULLS DISTINCT in unique constraints."""
        assert features_25.supports_nulls_distinct_unique_constraints is False

    def test_supports_temporal_subtraction(self, features_25):
        """Firebird uses DATEDIFF(), not - operator for temporal subtraction."""
        assert features_25.supports_temporal_subtraction is False

    def test_supports_unlimited_charfield(self, features_25):
        """Firebird requires VARCHAR length."""
        assert features_25.supports_unlimited_charfield is False

    def test_test_collations(self, features_25):
        """Test collations dict is defined."""
        assert features_25.test_collations is not None
        assert isinstance(features_25.test_collations, dict)
        assert "non_default" in features_25.test_collations

    def test_test_now_utc_template(self, features_25):
        """UTC timestamp template is defined."""
        assert features_25.test_now_utc_template is not None
        assert "TIMESTAMP" in features_25.test_now_utc_template



"""
Tests for DatabaseCreation class.
"""

from unittest.mock import MagicMock, patch

import pytest


class TestDatabaseCreation:
    """Test cases for DatabaseCreation methods."""

    @pytest.fixture
    def creation(self, mock_connection):
        """Create DatabaseCreation instance with mock connection."""
        from django_firebird.creation import DatabaseCreation

        return DatabaseCreation(mock_connection)

    # =========================================================================
    # _quote_name tests
    # =========================================================================

    def test_quote_name(self, creation):
        """Test _quote_name delegates to ops.quote_name."""
        creation.connection.ops.quote_name = MagicMock(return_value='"test"')

        result = creation._quote_name("test")

        assert result == '"test"'
        creation.connection.ops.quote_name.assert_called_once_with("test")

    # =========================================================================
    # _get_test_db_params tests
    # =========================================================================

    def test_get_test_db_params_defaults(self, creation):
        """Test default test database parameters."""
        creation.connection.settings_dict = {
            "USER": "SYSDBA",
            "PASSWORD": "masterkey",
            "HOST": None,
            "PORT": None,
            "OPTIONS": {},
            "TEST": {},
        }

        params = creation._get_test_db_params()

        assert params["user"] == "SYSDBA"
        assert params["password"] == "masterkey"
        assert params["page_size"] == 8192
        assert params["charset"] == "UTF8"

    def test_get_test_db_params_custom(self, creation):
        """Test custom test database parameters."""
        creation.connection.settings_dict = {
            "USER": "custom_user",
            "PASSWORD": "custom_pass",
            "HOST": "localhost",
            "PORT": "3050",
            "OPTIONS": {"charset": "WIN1252"},
            "TEST": {"PAGE_SIZE": 16384},
        }

        params = creation._get_test_db_params()

        assert params["user"] == "custom_user"
        assert params["password"] == "custom_pass"
        assert params["host"] == "localhost"
        assert params["port"] == "3050"
        assert params["page_size"] == 16384
        assert params["charset"] == "WIN1252"

    # =========================================================================
    # _get_test_db_name tests
    # =========================================================================

    def test_get_test_db_name_from_test_settings(self, creation):
        """Test getting test database name from TEST settings."""
        creation.connection.settings_dict = {
            "NAME": "/path/to/prod.fdb",
            "TEST": {"NAME": "/tmp/custom_test.fdb"},
        }

        result = creation._get_test_db_name()

        assert result == "/tmp/custom_test.fdb"

    def test_get_test_db_name_generated(self, creation):
        """Test generating test database name from production name."""
        creation.connection.settings_dict = {
            "NAME": "/path/to/mydb.fdb",
            "TEST": {},
        }

        result = creation._get_test_db_name()

        assert result == "/tmp/test_mydb.fdb"

    def test_get_test_db_name_no_extension(self, creation):
        """Test generating test database name without .fdb extension."""
        creation.connection.settings_dict = {
            "NAME": "/path/to/mydb",
            "TEST": {},
        }

        result = creation._get_test_db_name()

        assert result == "/tmp/test_mydb.fdb"

    # =========================================================================
    # _create_test_db tests
    # =========================================================================

    def test_create_test_db_keepdb(self, creation):
        """Test _create_test_db with keepdb=True."""
        creation.connection.settings_dict = {
            "NAME": "/path/to/prod.fdb",
            "TEST": {"NAME": "/tmp/test.fdb"},
        }

        with patch.object(creation, "_get_test_db_params"):
            result = creation._create_test_db(
                verbosity=0, autoclobber=False, keepdb=True
            )

        assert result == "/tmp/test.fdb"

    @patch("os.path.exists")
    @patch("os.remove")
    def test_create_test_db_existing_autoclobber(
        self, mock_remove, mock_exists, creation
    ):
        """Test _create_test_db removes existing database with autoclobber."""
        mock_exists.return_value = True
        creation.connection.settings_dict = {
            "NAME": "/path/to/prod.fdb",
            "USER": "SYSDBA",
            "PASSWORD": "masterkey",
            "HOST": None,
            "PORT": None,
            "OPTIONS": {},
            "TEST": {"NAME": "/tmp/test.fdb"},
        }

        with patch.object(creation, "_create_database"):
            creation._create_test_db(verbosity=0, autoclobber=True, keepdb=False)

        mock_remove.assert_called_once_with("/tmp/test.fdb")

    @patch("os.path.exists")
    def test_create_test_db_new_database(self, mock_exists, creation):
        """Test _create_test_db creates new database."""
        mock_exists.return_value = False
        creation.connection.settings_dict = {
            "NAME": "/path/to/prod.fdb",
            "USER": "SYSDBA",
            "PASSWORD": "masterkey",
            "HOST": None,
            "PORT": None,
            "OPTIONS": {},
            "TEST": {"NAME": "/tmp/test.fdb"},
        }

        with patch.object(creation, "_create_database") as mock_create:
            creation._create_test_db(verbosity=0, autoclobber=False, keepdb=False)

        mock_create.assert_called_once()

    # =========================================================================
    # _create_database tests
    # =========================================================================

    def test_create_database_local(self, creation):
        """Test creating local database."""
        with patch("firebird.driver.create_database") as mock_create:
            creation._create_database(
                "/tmp/test.fdb",
                {
                    "user": "SYSDBA",
                    "password": "masterkey",
                    "page_size": 8192,
                },
            )

        mock_create.assert_called_once()
        call_kwargs = mock_create.call_args[1]
        assert call_kwargs["database"] == "/tmp/test.fdb"
        assert call_kwargs["user"] == "SYSDBA"
        assert call_kwargs["password"] == "masterkey"
        assert call_kwargs["page_size"] == 8192

    def test_create_database_remote_with_host(self, creation):
        """Test creating remote database with host."""
        with patch("firebird.driver.create_database") as mock_create:
            creation._create_database(
                "/tmp/test.fdb",
                {
                    "host": "server",
                    "user": "SYSDBA",
                    "password": "masterkey",
                    "page_size": 8192,
                },
            )

        call_kwargs = mock_create.call_args[1]
        assert call_kwargs["database"] == "server:/tmp/test.fdb"

    def test_create_database_remote_with_host_and_port(self, creation):
        """Test creating remote database with host and port."""
        with patch("firebird.driver.create_database") as mock_create:
            creation._create_database(
                "/tmp/test.fdb",
                {
                    "host": "server",
                    "port": "3050",
                    "user": "SYSDBA",
                    "password": "masterkey",
                    "page_size": 8192,
                },
            )

        call_kwargs = mock_create.call_args[1]
        assert call_kwargs["database"] == "server/3050:/tmp/test.fdb"

    # =========================================================================
    # _destroy_test_db tests
    # =========================================================================

    @patch("os.path.exists")
    @patch("os.remove")
    def test_destroy_test_db(self, mock_remove, mock_exists, creation):
        """Test _destroy_test_db removes database file."""
        mock_exists.return_value = True
        creation.connection.close = MagicMock()

        creation._destroy_test_db("/tmp/test.fdb", verbosity=0)

        creation.connection.close.assert_called_once()
        mock_remove.assert_called_once_with("/tmp/test.fdb")

    @patch("os.path.exists")
    @patch("os.remove")
    def test_destroy_test_db_not_exists(self, mock_remove, mock_exists, creation):
        """Test _destroy_test_db when database doesn't exist."""
        mock_exists.return_value = False
        creation.connection.close = MagicMock()

        creation._destroy_test_db("/tmp/test.fdb", verbosity=0)

        creation.connection.close.assert_called_once()
        mock_remove.assert_not_called()

    # =========================================================================
    # _clone_test_db tests
    # =========================================================================

    def test_clone_test_db_not_supported(self, creation):
        """Test _clone_test_db raises NotImplementedError."""
        with pytest.raises(NotImplementedError) as exc_info:
            creation._clone_test_db(suffix="1", verbosity=0)

        assert "not supported for Firebird" in str(exc_info.value)



"""
Tests for DatabaseSchemaEditor class.
"""

from unittest.mock import MagicMock, patch

import pytest


class TestDatabaseSchemaEditor:
    """Test cases for DatabaseSchemaEditor methods."""

    @pytest.fixture
    def schema_editor(self, mock_connection):
        """Create DatabaseSchemaEditor instance with mock connection."""
        from django_firebird.schema import DatabaseSchemaEditor

        return DatabaseSchemaEditor(mock_connection)

    # =========================================================================
    # SQL template tests
    # =========================================================================

    def test_sql_create_index_template(self, schema_editor):
        """Test CREATE INDEX template includes condition placeholder."""
        assert "%(columns)s" in schema_editor.sql_create_index
        assert "%(condition)s" in schema_editor.sql_create_index

    def test_sql_create_unique_index_template(self, schema_editor):
        """Test CREATE UNIQUE INDEX template includes condition placeholder."""
        assert "%(columns)s" in schema_editor.sql_create_unique_index
        assert "%(condition)s" in schema_editor.sql_create_unique_index

    def test_sql_create_expression_index_template(self, schema_editor):
        """Test expression index uses COMPUTED BY syntax."""
        assert "COMPUTED BY" in schema_editor.sql_create_expression_index
        assert "%(expressions)s" in schema_editor.sql_create_expression_index
        assert "%(condition)s" in schema_editor.sql_create_expression_index

    def test_sql_create_unique_expression_index_template(self, schema_editor):
        """Test unique expression index uses COMPUTED BY syntax."""
        assert "COMPUTED BY" in schema_editor.sql_create_unique_expression_index
        assert "UNIQUE" in schema_editor.sql_create_unique_expression_index
        assert "%(expressions)s" in schema_editor.sql_create_unique_expression_index

    def test_sql_rename_column_template(self, schema_editor):
        """Test Firebird's column rename syntax (ALTER ... TO)."""
        assert "ALTER COLUMN" in schema_editor.sql_rename_column
        assert "TO" in schema_editor.sql_rename_column

    def test_sql_delete_index_template(self, schema_editor):
        """Test DROP INDEX template."""
        assert schema_editor.sql_delete_index == "DROP INDEX %(name)s"

    # =========================================================================
    # quote_value tests
    # =========================================================================

    def test_quote_value_string(self, schema_editor):
        """Test quoting string values."""
        assert schema_editor.quote_value("test") == "'test'"

    def test_quote_value_string_with_quotes(self, schema_editor):
        """Test quoting string with embedded single quotes."""
        assert schema_editor.quote_value("it's") == "'it''s'"

    def test_quote_value_boolean_true(self, schema_editor):
        """Test quoting boolean True."""
        assert schema_editor.quote_value(True) == "1"

    def test_quote_value_boolean_false(self, schema_editor):
        """Test quoting boolean False."""
        assert schema_editor.quote_value(False) == "0"

    def test_quote_value_none(self, schema_editor):
        """Test quoting None."""
        assert schema_editor.quote_value(None) == "NULL"

    def test_quote_value_integer(self, schema_editor):
        """Test quoting integer."""
        assert schema_editor.quote_value(42) == "42"

    def test_quote_value_float(self, schema_editor):
        """Test quoting float."""
        assert schema_editor.quote_value(3.14) == "3.14"

    # =========================================================================
    # Sequence and trigger SQL tests
    # =========================================================================

    def test_create_sequence_sql(self, schema_editor):
        """Test sequence creation SQL."""
        # Mock the ops methods
        schema_editor.connection.ops._get_generator_name = MagicMock(
            return_value="GEN_MY_TABLE_ID"
        )

        sql = schema_editor._create_sequence_sql("my_table", "id")
        assert "CREATE SEQUENCE" in sql

    def test_delete_sequence_sql(self, schema_editor):
        """Test sequence deletion SQL."""
        schema_editor.connection.ops._get_generator_name = MagicMock(
            return_value="GEN_MY_TABLE_ID"
        )

        sql = schema_editor._delete_sequence_sql("my_table", "id")
        assert "DROP SEQUENCE" in sql

    def test_create_trigger_sql(self, schema_editor):
        """Test auto-increment trigger creation SQL."""
        schema_editor.connection.ops._get_generator_name = MagicMock(
            return_value="GEN_MY_TABLE_ID"
        )
        schema_editor.connection.ops._get_trigger_name = MagicMock(
            return_value="TRG_MY_TABLE_ID"
        )

        sql = schema_editor._create_trigger_sql("my_table", "id")
        assert "CREATE TRIGGER" in sql
        assert "BEFORE INSERT" in sql
        assert "NEXT VALUE FOR" in sql
        assert "IS NULL" in sql

    def test_delete_trigger_sql(self, schema_editor):
        """Test trigger deletion SQL."""
        schema_editor.connection.ops._get_trigger_name = MagicMock(
            return_value="TRG_MY_TABLE_ID"
        )

        sql = schema_editor._delete_trigger_sql("my_table", "id")
        assert "DROP TRIGGER" in sql

    # =========================================================================
    # _alter_column_type_sql tests
    # =========================================================================

    def test_alter_column_type_sql(self, schema_editor):
        """Test ALTER COLUMN TYPE SQL generation."""
        mock_model = MagicMock()
        mock_model._meta.db_table = "test_table"

        old_field = MagicMock()
        new_field = MagicMock()
        new_field.column = "my_column"

        sql, params = schema_editor._alter_column_type_sql(
            mock_model, old_field, new_field, "VARCHAR(100)"
        )

        assert "ALTER TABLE" in sql
        assert "ALTER COLUMN" in sql
        assert "TYPE" in sql
        assert "VARCHAR(100)" in sql
        assert params == []

    # =========================================================================
    # _alter_column_null_sql tests
    # =========================================================================

    def test_alter_column_null_sql_set_nullable(self, schema_editor):
        """Test making column nullable."""
        mock_model = MagicMock()
        mock_model._meta.db_table = "test_table"

        old_field = MagicMock()
        new_field = MagicMock()
        new_field.column = "my_column"
        new_field.null = True

        sql, params = schema_editor._alter_column_null_sql(
            mock_model, old_field, new_field
        )

        assert "DROP NOT NULL" in sql
        assert params == []

    def test_alter_column_null_sql_set_not_nullable(self, schema_editor):
        """Test making column not nullable."""
        mock_model = MagicMock()
        mock_model._meta.db_table = "test_table"

        old_field = MagicMock()
        new_field = MagicMock()
        new_field.column = "my_column"
        new_field.null = False

        sql, params = schema_editor._alter_column_null_sql(
            mock_model, old_field, new_field
        )

        assert "SET NOT NULL" in sql
        assert params == []

    # =========================================================================
    # _alter_column_default_sql tests
    # =========================================================================

    def test_alter_column_default_sql_drop(self, schema_editor):
        """Test dropping column default."""
        mock_model = MagicMock()
        mock_model._meta.db_table = "test_table"

        old_field = MagicMock()
        new_field = MagicMock()
        new_field.column = "my_column"

        sql, params = schema_editor._alter_column_default_sql(
            mock_model, old_field, new_field, drop=True
        )

        assert "DROP DEFAULT" in sql
        assert params == []

    def test_alter_column_default_sql_set(self, schema_editor):
        """Test setting column default."""
        mock_model = MagicMock()
        mock_model._meta.db_table = "test_table"

        old_field = MagicMock()
        new_field = MagicMock()
        new_field.column = "my_column"

        with patch.object(
            schema_editor, "effective_default", return_value="default_value"
        ):
            sql, params = schema_editor._alter_column_default_sql(
                mock_model, old_field, new_field, drop=False
            )

        assert "SET DEFAULT" in sql
        assert "'default_value'" in sql
        assert params == []

    # =========================================================================
    # _field_should_be_altered tests
    # =========================================================================

    def test_field_should_be_altered_auto_field(self, schema_editor):
        """Test that AutoField changes are ignored."""
        old_field = MagicMock()
        old_field.get_internal_type.return_value = "AutoField"
        new_field = MagicMock()

        result = schema_editor._field_should_be_altered(old_field, new_field)

        assert result is False

    def test_field_should_be_altered_big_auto_field(self, schema_editor):
        """Test that BigAutoField changes are ignored."""
        old_field = MagicMock()
        old_field.get_internal_type.return_value = "BigAutoField"
        new_field = MagicMock()

        result = schema_editor._field_should_be_altered(old_field, new_field)

        assert result is False

    def test_field_should_be_altered_small_auto_field(self, schema_editor):
        """Test that SmallAutoField changes are ignored."""
        old_field = MagicMock()
        old_field.get_internal_type.return_value = "SmallAutoField"
        new_field = MagicMock()

        result = schema_editor._field_should_be_altered(old_field, new_field)

        assert result is False



"""
Tests for DatabaseIntrospection class.
"""

from unittest.mock import MagicMock

import pytest


class TestDatabaseIntrospection:
    """Test cases for DatabaseIntrospection methods."""

    @pytest.fixture
    def introspection(self, mock_connection):
        """Create DatabaseIntrospection instance with mock connection."""
        from django_firebird.introspection import DatabaseIntrospection

        return DatabaseIntrospection(mock_connection)

    # =========================================================================
    # Data types reverse mapping tests
    # =========================================================================

    def test_data_types_reverse_smallint(self, introspection):
        """Test SMALLINT maps to SmallIntegerField."""
        assert introspection.data_types_reverse[7] == "SmallIntegerField"

    def test_data_types_reverse_integer(self, introspection):
        """Test INTEGER maps to IntegerField."""
        assert introspection.data_types_reverse[8] == "IntegerField"

    def test_data_types_reverse_bigint(self, introspection):
        """Test BIGINT maps to BigIntegerField."""
        assert introspection.data_types_reverse[16] == "BigIntegerField"

    def test_data_types_reverse_float(self, introspection):
        """Test FLOAT maps to FloatField."""
        assert introspection.data_types_reverse[10] == "FloatField"

    def test_data_types_reverse_double(self, introspection):
        """Test DOUBLE PRECISION maps to FloatField."""
        assert introspection.data_types_reverse[27] == "FloatField"

    def test_data_types_reverse_date(self, introspection):
        """Test DATE maps to DateField."""
        assert introspection.data_types_reverse[12] == "DateField"

    def test_data_types_reverse_time(self, introspection):
        """Test TIME maps to TimeField."""
        assert introspection.data_types_reverse[13] == "TimeField"

    def test_data_types_reverse_timestamp(self, introspection):
        """Test TIMESTAMP maps to DateTimeField."""
        assert introspection.data_types_reverse[35] == "DateTimeField"

    def test_data_types_reverse_char(self, introspection):
        """Test CHAR maps to CharField."""
        assert introspection.data_types_reverse[14] == "CharField"

    def test_data_types_reverse_varchar(self, introspection):
        """Test VARCHAR maps to CharField."""
        assert introspection.data_types_reverse[37] == "CharField"

    def test_data_types_reverse_blob(self, introspection):
        """Test BLOB maps to TextField."""
        assert introspection.data_types_reverse[261] == "TextField"

    def test_data_types_reverse_boolean(self, introspection):
        """Test BOOLEAN maps to BooleanField (Firebird 3.0+)."""
        assert introspection.data_types_reverse[23] == "BooleanField"

    # =========================================================================
    # get_field_type tests
    # =========================================================================

    def test_get_field_type_basic(self, introspection):
        """Test basic field type lookup."""
        description = MagicMock()
        result = introspection.get_field_type(8, description)  # INTEGER
        assert result == "IntegerField"

    def test_get_field_type_blob_text(self, introspection):
        """Test BLOB with text subtype."""
        description = MagicMock()
        description.subtype = 1  # Text subtype

        result = introspection.get_field_type(261, description)
        assert result == "TextField"

    def test_get_field_type_blob_binary(self, introspection):
        """Test BLOB with binary subtype."""
        description = MagicMock()
        description.subtype = 0  # Binary subtype

        result = introspection.get_field_type(261, description)
        assert result == "BinaryField"

    # =========================================================================
    # get_table_list tests
    # =========================================================================

    def test_get_table_list_tables_only(self, introspection):
        """Test getting list of tables."""
        mock_cursor = MagicMock()
        mock_cursor.fetchall.return_value = [
            ("TABLE1", "t"),
            ("TABLE2", "t"),
        ]

        result = introspection.get_table_list(mock_cursor)

        assert len(result) == 2
        assert result[0].name == "TABLE1"
        assert result[0].type == "t"
        assert result[1].name == "TABLE2"

    def test_get_table_list_with_views(self, introspection):
        """Test getting list with views."""
        mock_cursor = MagicMock()
        mock_cursor.fetchall.return_value = [
            ("TABLE1", "t"),
            ("VIEW1", "v"),
        ]

        result = introspection.get_table_list(mock_cursor)

        assert len(result) == 2
        assert result[0].type == "t"
        assert result[1].type == "v"

    def test_get_table_list_empty(self, introspection):
        """Test empty table list."""
        mock_cursor = MagicMock()
        mock_cursor.fetchall.return_value = []

        result = introspection.get_table_list(mock_cursor)

        assert result == []

    # =========================================================================
    # get_table_description tests
    # =========================================================================

    def test_get_table_description_basic(self, introspection):
        """Test basic table description."""
        mock_cursor = MagicMock()
        # name, type, length, precision, scale, nullable, default, char_len, subtype, comment, field_source, collation
        mock_cursor.fetchall.return_value = [
            ("ID", 8, 4, None, None, 0, None, None, None, None, "RDB$1", None),
            (
                "NAME",
                37,
                100,
                None,
                None,
                1,
                None,
                100,
                None,
                "Customer name",
                "RDB$2",
                "UTF8",
            ),
        ]

        result = introspection.get_table_description(mock_cursor, "TEST")

        assert len(result) == 2
        assert result[0].name == "ID"
        assert result[0].type_code == 8
        assert result[0].null_ok is False
        assert result[1].name == "NAME"
        assert result[1].null_ok is True
        assert result[1].comment == "Customer name"
        assert result[1].collation == "UTF8"

    def test_get_table_description_varchar_length(self, introspection):
        """Test VARCHAR uses character length."""
        mock_cursor = MagicMock()
        # name, type, length, precision, scale, nullable, default, char_len, subtype, comment, field_source, collation
        mock_cursor.fetchall.return_value = [
            ("FIELD", 37, 400, None, None, 1, None, 100, None, None, "RDB$1", None),
        ]

        result = introspection.get_table_description(mock_cursor, "TEST")

        assert result[0].display_size == 100  # Should use char_length, not byte length

    def test_get_table_description_uppercase_table(self, introspection):
        """Test table name is uppercased in query."""
        mock_cursor = MagicMock()
        mock_cursor.fetchall.return_value = []

        introspection.get_table_description(mock_cursor, "lowercase_table")

        # Verify the table name was uppercased in the query
        call_args = mock_cursor.execute.call_args
        assert "LOWERCASE_TABLE" in call_args[0][1]

    # =========================================================================
    # get_primary_key_column tests
    # =========================================================================

    def test_get_primary_key_column_found(self, introspection):
        """Test finding primary key column."""
        mock_cursor = MagicMock()
        mock_cursor.fetchone.return_value = ("ID",)

        result = introspection.get_primary_key_column(mock_cursor, "TEST")

        assert result == "ID"

    def test_get_primary_key_column_not_found(self, introspection):
        """Test when no primary key exists."""
        mock_cursor = MagicMock()
        mock_cursor.fetchone.return_value = None

        result = introspection.get_primary_key_column(mock_cursor, "TEST")

        assert result is None

    # =========================================================================
    # get_relations tests
    # =========================================================================

    def test_get_relations_single_column_fk(self, introspection):
        """Test single-column foreign key."""
        mock_cursor = MagicMock()
        mock_cursor.fetchall.return_value = [
            ("COMPANY_ID", "COMPANY", "ID"),
        ]

        result = introspection.get_relations(mock_cursor, "CUSTOMER")

        assert "COMPANY_ID" in result
        assert result["COMPANY_ID"] == ("COMPANY", "ID")

    def test_get_relations_multiple_fks(self, introspection):
        """Test multiple foreign keys."""
        mock_cursor = MagicMock()
        mock_cursor.fetchall.return_value = [
            ("COMPANY_ID", "COMPANY", "ID"),
            ("CUSTOMER_ID", "CUSTOMER", "ID"),
            ("PRODUCT_ID", "PRODUCT", "ID"),
        ]

        result = introspection.get_relations(mock_cursor, "ORDER")

        assert len(result) == 3
        assert result["COMPANY_ID"] == ("COMPANY", "ID")
        assert result["CUSTOMER_ID"] == ("CUSTOMER", "ID")
        assert result["PRODUCT_ID"] == ("PRODUCT", "ID")

    def test_get_relations_no_fks(self, introspection):
        """Test table with no foreign keys."""
        mock_cursor = MagicMock()
        mock_cursor.fetchall.return_value = []

        result = introspection.get_relations(mock_cursor, "COMPANY")

        assert result == {}

    def test_get_relations_composite_fk(self, introspection):
        """Test composite foreign key (multiple columns)."""
        mock_cursor = MagicMock()
        # Composite FK: (ORDER_ID, PRODUCT_ID) -> ORDER_ITEM(ORDER_ID, PRODUCT_ID)
        mock_cursor.fetchall.return_value = [
            ("ORDER_ID", "ORDER", "ID"),
            ("LINE_NUMBER", "ORDER", "LINE_NUMBER"),
        ]

        result = introspection.get_relations(mock_cursor, "ORDER_DETAIL")

        # Each column gets its own mapping
        assert result["ORDER_ID"] == ("ORDER", "ID")
        assert result["LINE_NUMBER"] == ("ORDER", "LINE_NUMBER")

    # =========================================================================
    # get_constraints tests
    # =========================================================================

    def test_get_constraints_primary_key(self, introspection):
        """Test primary key constraint detection."""
        mock_cursor = MagicMock()

        def mock_execute(sql, params=None):
            if "PRIMARY KEY" in sql and "SELECT" in sql:
                # constraint_name, field_name, index_name, index_type
                mock_cursor.fetchall.return_value = [
                    ("PK_TEST", "ID", "PK_TEST_IDX", 0),  # 0 = ASC
                ]
            else:
                mock_cursor.fetchall.return_value = []

        mock_cursor.execute.side_effect = mock_execute

        result = introspection.get_constraints(mock_cursor, "TEST")

        assert "PK_TEST" in result
        assert result["PK_TEST"]["primary_key"] is True
        assert result["PK_TEST"]["unique"] is True
        assert "ID" in result["PK_TEST"]["columns"]
        assert result["PK_TEST"]["orders"] == ["ASC"]

    def test_get_constraints_unique(self, introspection):
        """Test unique constraint detection."""
        mock_cursor = MagicMock()

        def mock_execute(sql, params=None):
            if "PRIMARY KEY" in sql:
                mock_cursor.fetchall.return_value = []
            elif "UNIQUE" in sql and "RDB$CONSTRAINT_TYPE" in sql:
                # constraint_name, field_name, index_type
                mock_cursor.fetchall.return_value = [
                    ("UQ_EMAIL", "EMAIL", 0),  # 0 = ASC
                ]
            else:
                mock_cursor.fetchall.return_value = []

        mock_cursor.execute.side_effect = mock_execute

        result = introspection.get_constraints(mock_cursor, "TEST")

        assert "UQ_EMAIL" in result
        assert result["UQ_EMAIL"]["unique"] is True
        assert result["UQ_EMAIL"]["primary_key"] is False

    def test_get_constraints_foreign_key(self, introspection):
        """Test foreign key constraint detection."""
        mock_cursor = MagicMock()

        def mock_execute(sql, params=None):
            if "FOREIGN KEY" in sql:
                # constraint_name, field_name, ref_table, ref_field, index_type
                mock_cursor.fetchall.return_value = [
                    ("FK_CUSTOMER_COMPANY", "COMPANY_ID", "COMPANY", "ID", 0),
                ]
            else:
                mock_cursor.fetchall.return_value = []

        mock_cursor.execute.side_effect = mock_execute

        result = introspection.get_constraints(mock_cursor, "CUSTOMER")

        assert "FK_CUSTOMER_COMPANY" in result
        assert result["FK_CUSTOMER_COMPANY"]["foreign_key"] == ("COMPANY", "ID")
        assert result["FK_CUSTOMER_COMPANY"]["primary_key"] is False



"""
Firebird database creation for Django.

Handles creating and destroying test databases.
"""

import os

from django.db.backends.base.creation import BaseDatabaseCreation


class DatabaseCreation(BaseDatabaseCreation):
    """Firebird database creation class."""

    def _quote_name(self, name):
        """Quote a database name."""
        return self.connection.ops.quote_name(name)

    def _get_test_db_params(self):
        """Return parameters for test database creation."""
        settings = self.connection.settings_dict
        test_settings = settings.get("TEST", {})

        params = {
            "user": settings.get("USER", "SYSDBA"),
            "password": settings.get("PASSWORD", "masterkey"),
            "host": settings.get("HOST"),
            "port": settings.get("PORT"),
            "page_size": test_settings.get("PAGE_SIZE", 8192),
            "charset": settings.get("OPTIONS", {}).get("charset", "UTF8"),
        }

        return params

    def _get_test_db_name(self):
        """
        Return the name of the test database.
        """
        settings = self.connection.settings_dict
        test_settings = settings.get("TEST", {})

        if test_settings.get("NAME"):
            return test_settings["NAME"]

        # Generate test database name
        db_name = settings.get("NAME", "")
        base_name = os.path.splitext(os.path.basename(db_name))[0]
        return f"/tmp/test_{base_name}.fdb"

    def _create_test_db(self, verbosity, autoclobber, keepdb=False):
        """
        Create a test database.
        """
        test_db_name = self._get_test_db_name()
        params = self._get_test_db_params()

        if keepdb:
            return test_db_name

        # Remove existing database if exists
        if os.path.exists(test_db_name):
            if autoclobber:
                os.remove(test_db_name)
            else:
                confirm = input(
                    f"Database {test_db_name} already exists. "
                    "Type 'yes' to delete and recreate it: "
                )
                if confirm != "yes":
                    raise Exception("Database creation cancelled.")
                os.remove(test_db_name)

        # Create the database
        self._create_database(test_db_name, params)

        return test_db_name

    def _create_database(self, db_name, params):
        """Create a new Firebird database."""
        from firebird.driver import create_database

        dsn = db_name
        if params.get("host"):
            if params.get("port"):
                dsn = f"{params['host']}/{params['port']}:{db_name}"
            else:
                dsn = f"{params['host']}:{db_name}"

        create_database(
            database=dsn,
            user=params.get("user", "SYSDBA"),
            password=params.get("password", "masterkey"),
            page_size=params.get("page_size", 8192),
        )

    def _destroy_test_db(self, test_db_name, verbosity):
        """
        Destroy the test database.
        """
        # Close all connections
        self.connection.close()

        # Remove the database file
        if os.path.exists(test_db_name):
            os.remove(test_db_name)

    def _clone_test_db(self, suffix, verbosity, keepdb=False):
        """
        Clone the test database for parallel testing.
        Not supported for Firebird.
        """
        raise NotImplementedError("Cloning databases is not supported for Firebird.")
